#!/usr/bin/env python3

# sardinecake - manage VMs via OCI
#
# Copyright © 2024, IOhannes m zmölnig, forum::für::umläute
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


__version__ = "<unknown>"
__author__ = "IOhannes m zmölnig <zmoelnig@iem.at>"

from .auth import login, logout
from .exportvm import exportvm
from .importvm import importvm
from .ociregistry import fetch, push

try:
    from ._version import version as __version__
except ImportError:
    pass

__all__ = ["login", "logout", "exportvm", "importvm", "fetch", "push", "Domain", "QCOW"]
